/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <launchpad_hydra.h>

#include <stdio.h>
#include <dlfcn.h>
#include <cstring>
#include <fstream>
#include <limits.h>
#include <stdlib.h>

#include "log.h"
#include "launcher_env.h"

const char* __coreclr_lib = "/usr/share/dotnet.tizen/netcoreapp/libcoreclr.so";
const char* __dotnet_launcher = "/usr/bin/dotnet-launcher";

typedef int (*coreclr_preload_assembly_ptr)(const char* assemblyPath);
typedef int (*launcher_real_main_ptr)(int argc, char *argv[], const char* mode);

static std::string absolutePath(const std::string& path)
{
	std::string absPath;
	char realPath[PATH_MAX];
	if (realpath(path.c_str(), realPath) != nullptr && realPath[0] != '\0')
		absPath.assign(realPath);

	return absPath;
}

static void preloadAssemblies()
{
#ifdef USE_DEFAULT_BASE_ADDR
	putenv(const_cast<char *>("COMPlus_UseDefaultBaseAddr=1"));
#endif // USE_DEFAULT_BASE_ADDR

	void* coreclrLib = dlopen(__coreclr_lib, RTLD_NOW | RTLD_GLOBAL);
	if (coreclrLib == nullptr) {
		_DBG("dlopen failed to open libcoreclr.so");
		return;
	}

	coreclr_preload_assembly_ptr preloadAssembly;
	preloadAssembly = (coreclr_preload_assembly_ptr)dlsym(coreclrLib, "coreclr_preload_assembly");
	if (preloadAssembly == nullptr) {
		_DBG("coreclr_preload_assembly is not found in the libcoreclr.so");
		return;
	}

	std::ifstream preloadList(AOT_PRELOAD_PATH);
	if (preloadList) {
		std::string path;
		while (getline(preloadList, path)) {
			int st = preloadAssembly(absolutePath(path).c_str());
			if (st != 0) {
				_DBG("preload of %s failed! (0x%08x)", path.c_str(), st);
			} else {
				_DBG("preload of %s succeded", path.c_str());
			} 
		}
	}
}

int main(int argc, char** argv) 
{
	hydra_lifecycle_callback_s hydra_callback;

	hydra_callback.precreate = [](void* user_data) {
		_INFO("hydra : precreate");
		preloadAssemblies();
	};

	hydra_callback.create = [](void* user_data) {
		_INFO("hydra : create");
	};
	
	hydra_callback.fork = [](int argc, char **argv, void* user_data) -> int {
		_INFO("hydra : fork");
		void* launcher_h = dlopen(__dotnet_launcher, RTLD_NOW | RTLD_GLOBAL);
		if (launcher_h == nullptr) {
			_DBG("dlopen failed to open dotnet-launcher");
			return -1;
		}

		launcher_real_main_ptr realMain = (launcher_real_main_ptr)dlsym(launcher_h, "realMain");
		if (realMain == nullptr) {
			_DBG("realMain is not found in the dotnet-launcher");
			return -1;
		}

		return realMain(argc, argv, "default");			
	};

	hydra_callback.terminate = [](void* user_data)-> int {
		_INFO("hydra : terminate");
		return 0;
	};

	return launchpad_hydra_main(argc, argv, &hydra_callback, nullptr);
}

