/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "dotnet/dotnet_launcher.h"
#include "utils.h"
#include "log.h"

#include <cstdio>
#include <vector>
#include <memory>

#include <Ecore.h>
#include <Eina.h>
#include <aul.h>
#include <sys/types.h>
#include <unistd.h>

// By the specification, application id must be shorter than 50 character.
// To add margin, set CMD_LINE_SIZE to 64. (padding size is included)
#define APPID_MAX_LENGTH	52

static std::string StandaloneOption("--standalone");
static std::string PaddingOption("--PADDING_TO_CHANGE_CMDLINE");

extern "C" int realMain(int argc, char *argv[], const char* mode)
{
	int i;
	bool standaloneMode = false;
	char* standalonePath = nullptr;
	bool corerunMode = false;
	bool paddingExist = false;

	std::vector<char*> vargs;

	// start index 1 to avoid passing executable name "dotnet-launcher" as a parameter
	for (i = 1; i < argc; i++) {
		if (StandaloneOption.compare(argv[i]) == 0) {
			standaloneMode = true;

			if (i > argc - 1) {
				fprintf(stderr, "Assembly path must be after \"--standalone\" option\n");
				return 1;
			}
			i++;
			standalonePath = argv[i];
		} else if (PaddingOption.compare(argv[i]) == 0) {
			paddingExist = true;
		} else {
			vargs.push_back(argv[i]);
		}
	}

	if (isManagedAssembly(argv[1]) || isNativeImage(argv[1])) {
		corerunMode = true;
	}

	using tizen::runtime::Launchpad;
	using tizen::runtime::AppInfo;
	using tizen::runtime::dotnetcore::CoreRuntime;

	std::unique_ptr<CoreRuntime> runtime(new CoreRuntime(mode));

	// Intiailize ecore first (signal handlers, etc.) before runtime init.
	ecore_init();

	if (corerunMode) {
		_INFO("##### Run in corerun mode #####");
		char appId[APPID_MAX_LENGTH] = {0,};
		std::string appRoot;
		snprintf(appId, 16, "%s", "dotnet-launcher");
		appRoot = baseName(argv[1]);

		if (runtime->initialize(true) != 0) {
			_ERR("Failed to initialize");
			return 1;
		}

		int argsLen = vargs.size() - 1;
		char** args = &vargs[1];
		if (runtime->launch(appId, appRoot.c_str(), argv[1], argsLen, args)) {
			_ERR("Failed to launch");
			return 1;
		}
	} else if (standaloneMode) {
		_INFO("##### Run in standalone mode #####");
		char appId[APPID_MAX_LENGTH] = {0,};
		std::string appRoot;
		if (AUL_R_OK == aul_app_get_appid_bypid(getpid(), appId, sizeof(appId))) {
			const char* appRootPath = aul_get_app_root_path();
			if (appRootPath != nullptr)
				appRoot = std::string(appRootPath);
		} else {
			// If appId is not set, it is executed directly by cmdline.
			// In this case, appRoot is passed as an argument.
			snprintf(appId, 16, "%s", "dotnet-launcher");
			appRoot = baseName(baseName(standalonePath));
		}
		_INFO("AUL_APPID : %s", appId);

		if (runtime->initialize(true) != 0) {
			_ERR("Failed to initialize");
			return 1;
		}

		// change cmdline from dotnet-launcher to executable path
		int cmdlineSize = paddingExist ? APPID_MAX_LENGTH : APPID_MAX_LENGTH - PaddingOption.length();
		memset(argv[0], '\0', cmdlineSize);
		snprintf(argv[0], cmdlineSize, "%s", appId);

		int argsLen = vargs.size();
		char** args = &vargs[0];
		if (runtime->launch(appId, appRoot.c_str(), standalonePath, argsLen, args)) {
			_ERR("Failed to launch");
			return 1;
		}
	} else {
		// change cmdline from dotnet-hydra-launcher to dotnet-launcher
		if (strcmp(argv[0], "/usr/bin/dotnet-hydra-launcher") == 0) {
			memset(argv[0], '\0', strlen("/usr/bin/dotnet-hydra-launcher"));
			snprintf(argv[0], strlen("/usr/bin/dotnet-launcher") + 1, "/usr/bin/dotnet-launcher");
		}

		Launchpad.onCreate = [&runtime]() {
			if (runtime->initialize(false) != 0)
				_ERR("Failed to initialized");
			else
				_INFO("Success to initialized");
		};

		Launchpad.onTerminate = [&runtime](const AppInfo& appInfo, int argc, char** argv) {
			_INFO("launch request with app path : %s", appInfo.path.c_str());
			_INFO("appId : %s", appInfo.id.c_str());
			_INFO("pkg : %s", appInfo.pkg.c_str());

			// aul_get_app_root_path() can return NULL for error case.
			if (appInfo.root.empty()) {
				_ERR("Failed to launch. root path is set to NULL");
			} else {
				// The launchpad pass the name of exe file to the first argument.
				// For the C# spec, we have to skip this first argument.
				if (runtime->launch(appInfo.id.c_str(), appInfo.root.c_str(), appInfo.path.c_str(), argc-1, argv+1))
					_ERR("Failed to launch");
			}
		};
		int ret = Launchpad.loaderMain(argc, argv);
		if (ret < 0) {
			_ERR("fail to start loaderMain. candidate process is not created.");
			return 1;
		}
	}

	return 0;
}

int main(int argc, char *argv[])
{
	return realMain(argc, argv, "default");
}
