/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "utils.h"
#include "ni_common.h"
#include "tac_common.h"

#include <algorithm>
#include <cstdio>
#include <cstring>
#include <vector>

std::vector<std::string> getCmdArgs(char** begin, char** end)
{
	std::vector<std::string> list;
	for (char** itr = begin+1; itr != end; itr++) {
		if (strncmp(*itr, "--", 2) != 0) {
			list.push_back(*itr);
		}
	}
	return list;
}

static void help(const char *argv0)
{
	const char* helpDesc =
		"Usage: %s [OPTIONS] COMMAND <paths or pkg name>\n"
		"\n"
		"Commands:\n"
		"       --help                 - Display this screen\n"
		"       --ni-system            - Create NI under System DLLs\n"
		"       --ni-dll               - Create NI for DLL\n"
		"       --ni-pkg               - Create NI for package\n"
		"       --ni-pkg-dll           - Create NI for DLL in package\n"
		"       --ni-dir               - Create NI for directory\n"
		"       --ni-reset-system      - Remove System NI files\n"
		"       --ni-reset-pkg         - Remove App NI files\n"
		"       --ni-regen-all-app     - Re-generate All App NI files\n"
		"       --tac-regen-all        - Re-generate All TAC files\n"
		"       --tac-restore-db       - Restore TAC Database\n"
		"       --tac-disable-pkg      - Disable TAC for package\n"
		"       --tac-enable-pkg       - Enable TAC for package\n"
		"\n"
		"Options:\n"
		"       --r2r                  - Generate a Ready-To-Run image (disables /FragileNonVersionable)\n"
		"       --compatibility        - Compatibility mode for older versions of crossgen\n"
		"                                (replaces /r with /Trusted_Platform_Assemblies)\n"
		"       --verbose              - Display verbose information\n"
		"       --instrument           - Generate an instrumented image for profiling (enables /Tuning)\n"
		"\n"
		"Example:\n"
		"1. Create native image for dlls and exes under platform directories\n"
		"   # %s --ni-system\n"
		"2. Create native image for dll\n"
		"   # %s --ni-dll /usr/bin/Tizen.Runtime.dll\n"
		"3. Create native image under the package's bin and lib directory\n"
		"   # %s --ni-pkg org.tizen.FormsGallery\n"
		"4. Regenerate native images for all installed .net packages with ready-to-run option\n"
		"   # %s --r2r --ni-regen-all-app\n\n";
	printf(helpDesc, argv0, argv0, argv0, argv0, argv0);
}

int main(int argc, char* argv[])
{
	DWORD flags = 0;
	bool pkgMode = false;
	bool dllMode = false;
	bool dirMode = false;
	bool rmPkgMode = false;
	bool pkgDllMode = false;
	bool disableTacMode = false;
	bool enableTacMode = false;

	NiCommonOption option = {std::string(), std::string(), std::string()};
	if (initNICommon(&option) != NI_ERROR_NONE) {
		return -1;
	}

	// Parse optional switches first.
	if (cmdOptionExists(argv, argv+argc, "--r2r")) {
		flags |= NI_FLAGS_ENABLER2R;
	}
	if (cmdOptionExists(argv, argv+argc, "--compatibility")) {
		flags |= NI_FLAGS_COMPATIBILITY;
	}
	if (cmdOptionExists(argv, argv+argc, "--instrument")) {
		flags |= NI_FLAGS_INSTRUMENT;
	}
	if (cmdOptionExists(argv, argv+argc, "--verbose")) {
		flags |= NI_FLAGS_VERBOSE;
	}

	// The following commands are mutually exclusive.
	if (cmdOptionExists(argv, argv+argc, "--help")) {
		help(argv[0]);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-system")) {
		createNiPlatform(flags);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-dll")) {
		dllMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-pkg")) {
		pkgMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-dir")) {
		dirMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-reset-system")) {
		removeNiPlatform();
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-reset-pkg")) {
		rmPkgMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-pkg-dll")) {
		pkgDllMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--ni-regen-all-app")) {
		regenerateAppNI(flags);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--tac-regen-all")) {
		regenerateTACNI(flags);
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--tac-restore-db")) {
		restoreTACDB();
		return 0;
	} else if (cmdOptionExists(argv, argv+argc, "--tac-disable-pkg")) {
		disableTacMode = true;
	} else if (cmdOptionExists(argv, argv+argc, "--tac-enable-pkg")) {
		enableTacMode = true;
	} else {
		help(argv[0]);
		return 0;
	}

	std::vector<std::string> args = getCmdArgs(argv, argv+argc);

	if (args.size() < 1) {
		if (pkgMode || rmPkgMode || disableTacMode || enableTacMode) {
			fprintf(stderr, "Package name is missed\n");
		} else if (dllMode) {
			fprintf(stderr, "DLL path is missing.\n");
		}
		help(argv[0]);
		return -1;
	} else if (args.size() < 2) {
		if (pkgDllMode) {
			fprintf(stderr, "Package name or DLL path is missing.\n");
			help(argv[0]);
			return -1;
		}
	}

	if (pkgMode) {
		for (const std::string pkg : args) {
			// if there is AOTed dlls under package root, that is skiped.
			int ret = createNiUnderPkgRoot(pkg, flags);
			if (ret == NI_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to generate NI file [%s]\n", pkg.c_str());
				return -1;
			}
			ret = createTACPkgRoot(pkg, flags);
			if (ret == TAC_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != TAC_ERROR_NONE) {
				fprintf(stderr, "Failed to generate symbolic link file [%s]\n", pkg.c_str());
				return -1;
			}
		}
	} else if (pkgDllMode) {
		int ret = createNiDllUnderPkgRoot(args[0], args[1], flags);
		if (ret == NI_ERROR_INVALID_PACKAGE) {
			fprintf(stderr, "Failed to get root path from [%s]\n", args[0].c_str());
			return -1;
		} else if (ret == NI_ERROR_ALREADY_EXIST) {
			// skip for already exist case
			return -1;
		} else if (ret != NI_ERROR_NONE) {
			fprintf(stderr, "Failed to generate NI file [%s]\n", args[1].c_str());
			return -1;
		}
	} else if (rmPkgMode) {
		for (const std::string pkg : args) {
			int ret = removeNiUnderPkgRoot(pkg);
			if (ret == NI_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to remove dlls for given package [%s]\n", pkg.c_str());
				return -1;
			}
			ret = resetTACPackage(pkg);
			if (ret == TAC_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != TAC_ERROR_NONE) {
				fprintf(stderr, "Failed to remove symlink for given package [%s]\n", pkg.c_str());
				return -1;
			}
		}
	} else if (dllMode) {
		// donot return error code for generation failure.
		// we have to run crossgen for all input dlls.
		for (const std::string dll : args) {
			int ret = createNiDll(dll, flags);
			if (ret == NI_ERROR_ALREADY_EXIST) {
				// skip for already exist case
			} else if (ret != NI_ERROR_NONE) {
				fprintf(stderr, "Failed to generate NI file [%s]\n", dll.c_str());
			}
		}
	} else if (dirMode) {
		createNiUnderDirs(args.data(), args.size(), flags);
	} else if (disableTacMode) {
		for (const std::string pkg : args) {
			int ret = disableTACPackage(pkg);
			if (ret == TAC_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != TAC_ERROR_NONE) {
				fprintf(stderr, "Failed to disable tac [%s]\n", pkg.c_str());
				return -1;
			}
		}
	} else if (enableTacMode) {
		for (const std::string pkg : args) {
			int ret = enableTACPackage(pkg);
			if (ret == TAC_ERROR_INVALID_PACKAGE) {
				fprintf(stderr, "Failed to get root path from [%s]\n", pkg.c_str());
				return -1;
			} else if (ret != TAC_ERROR_NONE) {
				fprintf(stderr, "Failed to enable tac [%s]\n", pkg.c_str());
				return -1;
			}
		}
	}
	return 0;
}
