/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <fstream>
#include <json/json.h>
#include <pkgmgr-info.h>
#include <pkgmgr_installer_info.h>

#include "log.h"
#include "utils.h"
#include "tac_common.h"
#include "db_manager.h"
#include "path_manager.h"

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "DOTNET_INSTALLER_PLUGIN"

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)
static const char* __TAC_DIR = __STR(TAC_DIR);
#undef __STR
#undef __XSTR

static sqlite3 *tac_db = NULL;
std::vector<std::string> restore_nuget;

static void cleanupDirectory()
{
	std::vector<std::string> removeNuget;
	try {
		for (auto& nuget : bf::recursive_directory_iterator(__TAC_DIR)) {
			bool isExist = false;
			std::string nugetPath = nuget.path().string();
			for (auto& restore : restore_nuget) {
				if (!bf::is_directory(nugetPath)) {
					isExist = true;
				}
				if (!strcmp(nugetPath.c_str(), restore.c_str()) ||
					!strcmp(nugetPath.c_str(), restore.substr(0, restore.rfind('/')).c_str())) {
					isExist = true;
					break;
				}
			}
			if (!isExist) {
				removeNuget.push_back(nugetPath);
			}
		}

		for (auto& rm : removeNuget) {
			if (!removeAll(rm)) {
				fprintf(stderr, "Failed to remove of %s\n", rm.c_str());
			}
		}
		removeNuget.clear();
	} catch (const bf::filesystem_error& error) {
		fprintf(stderr, "Failed to recursive directory: %s\n", error.what());
		return;
	}
}

// callback function of "pkgmgrinfo_appinfo_metadata_filter_foreach"
static int restoreDBCb(pkgmgrinfo_appinfo_h handle, void *userData)
{
	char *pkgId = NULL;
	char *root = NULL;
	char *exec = NULL;
	std::string rootPath;
	std::string execName;

	int ret = pkgmgrinfo_appinfo_get_pkgid(handle, &pkgId);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed to get pkgid\n");
		return -1;
	}

	enableTACPackage(std::string(pkgId));

	ret = pkgmgrinfo_appinfo_get_root_path(handle, &root);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed to get root path\n");
		return -1;
	}
	rootPath = root;

	ret = pkgmgrinfo_appinfo_get_exec(handle, &exec);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed to get exec name\n");
		return -1;
	}
	execName = std::string(exec).substr(std::string(exec).rfind('/') + 1);

	std::vector<std::string> parserData;
	std::string binDir = concatPath(rootPath, "bin");
	std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
	for (auto& npAssembly : depsJsonParser(rootPath, execName, getTPA())) {
		std::string nugetPackage = npAssembly.substr(0, npAssembly.rfind(':'));
		std::string assemblyName = npAssembly.substr(npAssembly.rfind(':') + 1);
		if (bf::exists(tacDir) && bf::exists(concatPath(tacDir, assemblyName))) {
			parserData.push_back(nugetPackage);
		}
	}
	std::sort(parserData.begin(), parserData.end());
	parserData.erase(unique(parserData.begin(), parserData.end()), parserData.end());

	for (auto& nuget : parserData) {
		if (tac_db) {
			std::string name = nuget.substr(0, nuget.find('/'));
			std::string version = nuget.substr(nuget.rfind('/') + 1);
			char *sql = sqlite3_mprintf(
				"INSERT INTO TAC (PKGID, NUGET, NAME, VERSION) " \
				"VALUES (%Q, %Q, %Q, %Q);",	pkgId, nuget.c_str(), name.c_str(), version.c_str());
			dbInsert(tac_db, TAC_APP_LIST_RESTORE_DB, sql);
			restore_nuget.push_back(concatPath(__TAC_DIR, nuget));
			sqlite3_free(sql);
		}
	}
	parserData.clear();
	return 0;
}

tac_error_e restoreTACDB()
{
	if (!removeFile(TAC_APP_LIST_RESTORE_DB)) {
		fprintf(stderr, "Failed to remove of %s\n", TAC_APP_LIST_RESTORE_DB);
		return TAC_ERROR_UNKNOWN;
	}

	std::string dbRestoreJournal = TAC_APP_LIST_RESTORE_DB + std::string("-journal");
	if (!removeFile(dbRestoreJournal)) {
		fprintf(stderr, "Failed to remove of %s\n", dbRestoreJournal.c_str());
		return TAC_ERROR_UNKNOWN;
	}

	if (initializePathManager(std::string(), std::string(), std::string())) {
		fprintf(stderr, "Fail to initialize PathManger\n");
		return TAC_ERROR_UNKNOWN;
	}

	tac_db = dbCreate(TAC_APP_LIST_RESTORE_DB);
	if (!tac_db) {
		fprintf(stderr, "Sqlite create error\n");
		return TAC_ERROR_UNKNOWN;
	}

	pkgmgrinfo_appinfo_metadata_filter_h handle;
	int ret = pkgmgrinfo_appinfo_metadata_filter_create(&handle);
	if (ret != PMINFO_R_OK) {
		return TAC_ERROR_UNKNOWN;
	}

	ret = pkgmgrinfo_appinfo_metadata_filter_add(handle, TAC_METADATA_KEY, METADATA_VALUE);
	if (ret != PMINFO_R_OK) {
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return TAC_ERROR_UNKNOWN;
	}

	ret = pkgmgrinfo_appinfo_metadata_filter_foreach(handle, restoreDBCb, NULL);
	if (ret != PMINFO_R_OK) {
		fprintf(stderr, "Failed pkgmgrinfo_appinfo_metadata_filter_foreach\n");
		pkgmgrinfo_appinfo_metadata_filter_destroy(handle);
		return TAC_ERROR_UNKNOWN;
	}
	fprintf(stderr, "Success pkgmgrinfo_appinfo_metadata_filter_foreach\n");

	pkgmgrinfo_appinfo_metadata_filter_destroy(handle);

	if (tac_db) {
		dbClose(tac_db);
		tac_db = NULL;
	}

	uid_t g_uid = 301; // app_fw
	gid_t g_gid = 301; // app_fw

	if (!copyFile(TAC_APP_LIST_RESTORE_DB, TAC_APP_LIST_DB)) {
		fprintf(stderr, "Failed to copy of %s\n", TAC_APP_LIST_DB);
		return TAC_ERROR_UNKNOWN;
	}
	if (!removeFile(TAC_APP_LIST_RESTORE_DB)) {
		fprintf(stderr, "Failed to remove of %s\n", TAC_APP_LIST_RESTORE_DB);
		return TAC_ERROR_UNKNOWN;
	}
	if (chown(TAC_APP_LIST_DB, g_uid, g_gid) == -1) {
		fprintf(stderr, "Failed to change owner and group name\n");
	}

	std::string dbJournal = TAC_APP_LIST_DB + std::string("-journal");
	if (!copyFile(dbRestoreJournal, dbJournal)) {
		fprintf(stderr, "Failed to copy of %s\n", dbJournal.c_str());
		return TAC_ERROR_UNKNOWN;
	}
	if (!removeFile(dbRestoreJournal)) {
		fprintf(stderr, "Failed to remove of %s\n", dbRestoreJournal.c_str());
		return TAC_ERROR_UNKNOWN;
	}
	if (chown(dbJournal.c_str(), g_uid, g_gid) == -1) {
		fprintf(stderr, "Failed to change owner and group name\n");
	}

	cleanupDirectory();

	return TAC_ERROR_NONE;
}

tac_error_e resetTACPackage(const std::string& pkgId)
{
	std::string pkgRoot;
	if (getRootPath(pkgId, pkgRoot) < 0) {
		return TAC_ERROR_INVALID_PACKAGE;
	}

	std::vector<std::string> tacNativeImage;
	std::string binDir = concatPath(pkgRoot, "bin");
	std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
	if (bf::exists(tacDir)) {
		try {
			for (auto& symlinkAssembly : bf::recursive_directory_iterator(tacDir)) {
				std::string symPath = symlinkAssembly.path().string();
				if (bf::is_symlink(symPath)) {
					if (isNativeImage(symPath)) {
						tacNativeImage.push_back(symPath);
					}
				}
			}
			for (auto& path : tacNativeImage) {
				if (!removeFile(path)) {
					fprintf(stderr, "Failed to remove of %s\n", path.c_str());
					return TAC_ERROR_UNKNOWN;
				}
			}
			tacNativeImage.clear();
		} catch (const bf::filesystem_error& error) {
			fprintf(stderr, "Failed to recursive directory: %s\n", error.what());
			return TAC_ERROR_UNKNOWN;
		}
	}
	return TAC_ERROR_NONE;
}

tac_error_e disableTACPackage(const std::string& pkgId)
{
	std::string pkgRoot;
	if (getRootPath(pkgId, pkgRoot) < 0) {
		return TAC_ERROR_INVALID_PACKAGE;
	}

	std::string binDir = concatPath(pkgRoot, "bin");
	std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
	std::string binNIDir = concatPath(binDir, APP_NI_SUB_DIR);
	if (bf::exists(tacDir)) {
		try {
			for (auto& symlinkAssembly : bf::recursive_directory_iterator(tacDir)) {
				std::string symPath = symlinkAssembly.path().string();
				std::string fileName = symlinkAssembly.path().filename().string();
				if (bf::is_symlink(symPath)) {
					std::string originPath = bf::read_symlink(symPath).string();
					if (!isNativeImage(symPath)) {
						std::string dllPath = concatPath(binDir, fileName);
						if (!copyFile(originPath, dllPath)) {
							fprintf(stderr, "Failed to copy of %s\n", dllPath.c_str());
							return TAC_ERROR_UNKNOWN;
						}
						updateAssemblyInfo(binDir.c_str(), concatPath(binDir, fileName).c_str());
					} else {
						std::string niPath = concatPath(binNIDir, fileName);
						if (!copyFile(originPath, niPath)) {
							fprintf(stderr, "Failed to copy of %s\n", niPath.c_str());
							return TAC_ERROR_UNKNOWN;
						}
						updateAssemblyInfo(binDir.c_str(), niPath.c_str());
					}
				}
			}
			if (!removeAll(tacDir)) {
				fprintf(stderr, "Failed to remove of %s\n", tacDir.c_str());
				return TAC_ERROR_UNKNOWN;
			}
		} catch (const bf::filesystem_error& error) {
			fprintf(stderr, "Failed to recursive directory: %s\n", error.what());
			return TAC_ERROR_UNKNOWN;
		}
	}
	return TAC_ERROR_NONE;
}

tac_error_e enableTACPackage(const std::string& pkgId)
{
	std::string rootPath;
	if (getRootPath(pkgId, rootPath) < 0) {
		return TAC_ERROR_INVALID_PACKAGE;
	}

	std::string execName;
	if (getExecName(pkgId, execName) < 0) {
		return TAC_ERROR_INVALID_PACKAGE;
	}

	std::string metaValue;
	if (getMetadataValue(pkgId, TAC_METADATA_KEY, metaValue) < 0) {
		return TAC_ERROR_INVALID_PACKAGE;
	}

	if (initializePathManager(std::string(), std::string(), std::string())) {
		fprintf(stderr, "Fail to initialize PathManger\n");
		return TAC_ERROR_UNKNOWN;
	}

	if (!strcmp(metaValue.c_str(), "true")) {
		std::string binDir = concatPath(rootPath, "bin");
		std::string tacDir = concatPath(binDir, TAC_SYMLINK_SUB_DIR);
		std::string binNIDir = concatPath(binDir, APP_NI_SUB_DIR);
		if (!bf::exists(tacDir)) {
			if (!createDir(tacDir)) {
				fprintf(stderr, "Cannot create directory: %s\n", tacDir.c_str());
				return TAC_ERROR_UNKNOWN;
			}
			updateAssemblyInfo(binDir.c_str(), tacDir.c_str());

			std::vector<std::string> enableNuget;
			for (auto& npAssembly : depsJsonParser(rootPath, execName, getTPA())) {
				std::string nugetPackage = npAssembly.substr(0, npAssembly.rfind(':'));
				std::string assemblyName = npAssembly.substr(npAssembly.rfind(':') + 1);
				std::string nugetPath = concatPath(__TAC_DIR, nugetPackage);
				if (bf::exists(nugetPath)) {
					std::string originPath = concatPath(nugetPath, assemblyName);
					if (bf::exists(originPath)) {
						enableNuget.push_back(originPath);
					}
				}
			}

			for (auto& originPath : enableNuget) {
				if (bf::exists(originPath)) {
					std::string fileName = originPath.substr(originPath.rfind('/') + 1);
					std::string NIFileName = fileName.substr(0, fileName.rfind(".dll")) + ".ni.dll";
					if (bf::exists(binNIDir)) {
						std::string originNIPath = originPath.substr(0, originPath.rfind(".dll")) + ".ni.dll";
						if (bf::exists(originNIPath)) {
							bf::create_symlink(originNIPath, concatPath(tacDir, NIFileName));
							fprintf(stderr, "%s symbolic link file generated successfully.\n", concatPath(tacDir, NIFileName).c_str());
							updateAssemblyInfo(tacDir.c_str(), concatPath(tacDir, NIFileName).c_str(), true);

							if (!removeFile(concatPath(binNIDir, NIFileName))) {
								fprintf(stderr, "Failed to remove of %s\n", concatPath(binNIDir, NIFileName).c_str());
								return TAC_ERROR_UNKNOWN;
							}
						}
					}
					bf::create_symlink(originPath, concatPath(tacDir, fileName));
					fprintf(stderr, "%s symbolic link file generated successfully.\n", concatPath(tacDir, fileName).c_str());
					updateAssemblyInfo(tacDir.c_str(), concatPath(tacDir, fileName).c_str(), true);

					if (!removeFile(concatPath(binDir, fileName))) {
						fprintf(stderr, "Failed to remove of %s\n", concatPath(binDir, fileName).c_str());
						return TAC_ERROR_UNKNOWN;
					}
				}
			}
			if (enableNuget.empty()) {
				if (!removeAll(tacDir)) {
					fprintf(stderr, "Failed to remove of %s\n", tacDir.c_str());
				}
			}
			enableNuget.clear();
		}
	} else {
		fprintf(stderr, "The metadata key is missing or the metadata value is false of [%s]\n", pkgId.c_str());
	}
	return TAC_ERROR_NONE;
}

//Parser the .deps.json file to get nuget information.
std::vector<std::string> depsJsonParser(std::string rootPath, std::string execName, std::string tpaList)
{
	std::vector<std::string> tpaAssemblies;
	splitPath(tpaList, tpaAssemblies);

	std::vector<std::string> parserData;
	std::string depsJsonName = execName.substr(0, execName.rfind(".dll")) + ".deps.json";
	std::string depsJsonPath = concatPath(rootPath, depsJsonName);
	try {
		if (bf::exists(depsJsonPath)) {
			std::ifstream ifs(depsJsonPath);
			Json::CharReaderBuilder reader;
			Json::Value root;
			std::string error;
			if (ifs.is_open()) {
				if (!Json::parseFromStream(reader, ifs, &root, &error)) {
					_ERR("Failed to parse of deps.json");
					ifs.close();
					tpaAssemblies.clear();
					return parserData;
				}
				const Json::Value runtimeTargetName = root["runtimeTarget"]["name"];
				const Json::Value nugetPackages = root["targets"][runtimeTargetName.asString().c_str()];
				std::vector<std::string> appDependencies;
				for (auto& nuget : nugetPackages.getMemberNames()) {
					if (strstr(nuget.c_str(), (execName.substr(0, execName.find(".Tizen."))).c_str()) != NULL ||
						strstr(nuget.c_str(), (execName.substr(0, execName.find(".dll"))).c_str()) != NULL) {
						const Json::Value assemblies = nugetPackages[nuget.c_str()]["runtime"];
						if (assemblies != Json::nullValue) {
							const Json::Value dependencies = nugetPackages[nuget.c_str()]["dependencies"];
							for (auto& dependency : dependencies.getMemberNames()) {
								appDependencies.push_back(dependency);
							}
						}
					}
				}
				for (auto& nuget : nugetPackages.getMemberNames()) {
					//Skip the nuget package related to Tizen
					if (strstr(nuget.c_str(), TIZEN_DOTNET_NUGET) == NULL &&
						strstr(nuget.c_str(), TIZEN_DOTNET_SDK_NUGET) == NULL &&
						strstr(nuget.c_str(), (execName.substr(0, execName.find(".Tizen."))).c_str()) == NULL &&
						strstr(nuget.c_str(), (execName.substr(0, execName.find(".dll"))).c_str()) == NULL) {
						const Json::Value assemblies = nugetPackages[nuget.c_str()]["runtime"];
						if (assemblies != Json::nullValue) {
							const Json::Value dependencies = nugetPackages[nuget.c_str()]["dependencies"];
							bool hasDependency = false;
							for (auto& dependency : dependencies.getMemberNames()) {
								//Skip the nugget package that is dependent on another nuget package
								if (strstr(dependency.c_str(), TIZEN_DOTNET_NUGET) == NULL &&
									strstr(dependency.c_str(), NET_STANDARD_LIBRARY_NUGET) == NULL) {
									hasDependency = true;
									for (auto& ad : appDependencies) {
										if (!strcmp(ad.c_str(), dependency.c_str())) {
											hasDependency = true;
											break;
										} else {
											hasDependency = false;
										}
									}
									if (hasDependency) break;
								}
							}
							if (!hasDependency) {
								bool isExistTpaAssembly = false;
								for (auto& assembly : assemblies.getMemberNames()) {
									std::string assemblyName = assembly.substr(assembly.rfind('/') + 1);
									//Skip the assembly present in the TPA list
									for (auto& tpa : tpaAssemblies) {
										if (!strcmp(replaceAll(tpa, ".ni.dll", ".dll").c_str(), assembly.c_str())) {
											isExistTpaAssembly = true;
											break;
										}
									}
									if (isExistTpaAssembly)	break;
								}
								if (!isExistTpaAssembly) {
									for (auto& assembly : assemblies.getMemberNames()) {
										std::string assemblyName = assembly.substr(assembly.rfind('/') + 1);
										parserData.push_back(nuget + ":" + assemblyName);
										_INFO("Nuget : [%s] / Assembly : [%s]", nuget.c_str(), assemblyName.c_str());
									}
								}
							}
						}
					}
				}
				appDependencies.clear();
				ifs.close();
			}
		}
	} catch (const Json::LogicError& error) {
		_ERR("Failed to parse Json: %s", error.what());
	}
	tpaAssemblies.clear();
	return parserData;
}
