/*
 * Copyright (c) 2019 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <algorithm>

#include "db_manager.h"
#include "log.h"
#include "tac_common.h"

#ifdef  LOG_TAG
#undef  LOG_TAG
#endif
#define LOG_TAG "DOTNET_INSTALLER_PLUGIN"

static int dbIntegrityCheckCb(void *user_data, int argc, char **argv, char **not_used)
{
	bool* is_db_corrupted = static_cast<bool*>(user_data);
	std::string result_DB_check = argv[0];
	std::transform(result_DB_check.begin(), result_DB_check.end(), result_DB_check.begin(), ::tolower);
	if(result_DB_check != "ok") {
		*is_db_corrupted = true;
		_ERR("DB integrity check failed");
		return -1;
	}
	_INFO("DB integrity result : %s", argv[0]);
	return 0;
}

sqlite3* dbCreate(std::string path)
{
	sqlite3 *sqlite = NULL;
	bool is_db_corrupted = false;
	int ret = sqlite3_open(path.c_str(), &sqlite);
	if (ret != SQLITE_OK) {
		_INFO("DB file is corrupted, start to recover corrupted db");
		if (restoreTACDB() != TAC_ERROR_NONE) {
			_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
			return NULL;
		}
		return dbCreate(path);
	}
	ret = sqlite3_exec(sqlite, "PRAGMA journal_mode = PERSIST", NULL, NULL, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
		sqlite3_close(sqlite);
		sqlite = NULL;
		return sqlite;
	}
	ret = sqlite3_exec(sqlite, CREATE_TAC_DB_TABLE, NULL, NULL, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
		sqlite3_close(sqlite);
		sqlite = NULL;
		return sqlite;
	}
	ret = sqlite3_exec(sqlite, "PRAGMA integrity_check", dbIntegrityCheckCb, &is_db_corrupted, NULL);
	if (ret == SQLITE_CORRUPT || is_db_corrupted) {
		_INFO("DB file is corrupted, start to recover corrupted db");
		if (restoreTACDB() != TAC_ERROR_NONE) {
			_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
			sqlite3_close(sqlite);
			sqlite = NULL;
			return sqlite;
		}
		return dbCreate(path);
	}
	return sqlite;
}

sqlite3* dbOpen(std::string path)
{
	sqlite3 *sqlite = NULL;
	bool is_db_corrupted = false;
	int ret = sqlite3_open(path.c_str(), &sqlite);
	if (ret != SQLITE_OK) {
		_INFO("DB file is corrupted, start to recover corrupted db");
		if (restoreTACDB() != TAC_ERROR_NONE) {
			_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
			return NULL;
		}
		return dbOpen(path);
	}
	ret = sqlite3_exec(sqlite, "PRAGMA integrity_check", dbIntegrityCheckCb, &is_db_corrupted, NULL);
	if (ret == SQLITE_CORRUPT || is_db_corrupted) {
		_INFO("DB file is corrupted, start to recover corrupted db");
		if (restoreTACDB() != TAC_ERROR_NONE) {
			_ERR("Sqlite error : [%d, %s]", ret, path.c_str());
			sqlite3_close(sqlite);
			sqlite = NULL;
			return sqlite;
		}
		return dbOpen(path);
	}
	return sqlite;
}

void dbFinalize(sqlite3_stmt *stmt)
{
	if (stmt) {
		sqlite3_finalize(stmt);
		stmt = NULL;
	}
}

void dbClose(sqlite3 *tac_db)
{
	if (tac_db) {
		sqlite3_exec(tac_db, "COMMIT;", NULL, NULL, NULL);
		sqlite3_close(tac_db);
		tac_db = NULL;
	}
}

void dbRollback(sqlite3 *tac_db)
{
	if (tac_db) {
		sqlite3_exec(tac_db, "ROLLBACK;", NULL, NULL, NULL);
		sqlite3_close(tac_db);
		tac_db = NULL;
	}
}

bool dbUpdate(sqlite3 *tac_db, std::string path, std::string query)
{
	sqlite3_stmt *stmt = NULL;
	int ret = sqlite3_exec(tac_db, "BEGIN;", NULL, NULL, NULL);
	ret = sqlite3_prepare(tac_db, query.c_str(), QUERY_MAX_LEN , &stmt, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbClose(tac_db);
		return false;
	}
	ret = sqlite3_step(stmt);
	if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbFinalize(stmt);
		dbClose(tac_db);
		return false;
	}
	dbFinalize(stmt);
	return true;
}

bool dbInsert(sqlite3 *tac_db, std::string path, std::string query)
{
	sqlite3_stmt *stmt = NULL;
	int ret = sqlite3_exec(tac_db, "BEGIN;", NULL, NULL, NULL);
	ret = sqlite3_prepare(tac_db, query.c_str(), QUERY_MAX_LEN , &stmt, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbClose(tac_db);
		return false;
	}
	ret = sqlite3_step(stmt);
	if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbFinalize(stmt);
		dbClose(tac_db);
		return false;
	}
	dbFinalize(stmt);
	return true;
}

std::vector<std::string> dbSelect(sqlite3 *tac_db, std::string path, std::string query)
{
	std::vector<std::string> updateDB;
	sqlite3_stmt* stmt = NULL;
	const char* str = NULL;
	int ret = sqlite3_prepare_v2(tac_db, query.c_str(), strlen(query.c_str()), &stmt, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbClose(tac_db);
		return updateDB;
	}
	while (sqlite3_step(stmt) == SQLITE_ROW) {
		str = (const char *) sqlite3_column_text(stmt, 2);
		updateDB.push_back((!str || !strlen(str)) ? NULL : strdup(str));
	}
	dbFinalize(stmt);
	return updateDB;
}

bool dbDelete(sqlite3 *tac_db, std::string path, std::string query)
{
	sqlite3_stmt *stmt = NULL;
	int ret = sqlite3_exec(tac_db, "BEGIN;", NULL, NULL, NULL);
	ret = sqlite3_prepare(tac_db, query.c_str(), QUERY_MAX_LEN , &stmt, NULL);
	if (ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbClose(tac_db);
		return false;
	}
	ret = sqlite3_step(stmt);
	if (ret != SQLITE_DONE && ret != SQLITE_ROW && ret != SQLITE_OK) {
		_ERR("Sqlite error : [%d, %s]", ret, sqlite3_errmsg(tac_db));
		dbFinalize(stmt);
		dbClose(tac_db);
		return false;
	}
	dbFinalize(stmt);
	return true;
}
