/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <vector>
#include <iterator>
#include <sstream>
#include <vconf.h>
#include <fstream>

#include "utils.h"
#include "plugin_manager.h"
#include "log.h"

static const char* __TIZEN_API_PATH_KEY = "db/dotnet/tizen_api_path";

#define __XSTR(x) #x
#define __STR(x) __XSTR(x)
static const char* __DEVICE_API_DIR = __STR(DEVICE_API_DIR);
static const char* __RUNTIME_DIR = __STR(RUNTIME_DIR);
#undef __STR
#undef __XSTR

typedef struct DllPath {
	std::string runtime_dir;
	std::string tizenfx_dir;
	std::string tizenfx_ref_dir;
	std::vector <std::string>extra_dirs;
} DllPath;

static DllPath* __dllPath = nullptr;
static std::string __tpa;
bool initializedPathManager = false;

// on success, return 0. otherwise return -1.
int initializePathManager(const std::string& runtimeDir, const std::string& tizenFXDir, const std::string& extraDir)
{
	if (initializedPathManager) {
		_INFO("Path manager already initialized");
		return 0;
	}

	__dllPath = new DllPath();
	if (!__dllPath) {
		_ERR("fail to allocate memory for dll path structure\n");
		return -1;
	}

	if (!runtimeDir.empty()) {
		__dllPath->runtime_dir = absolutePath(runtimeDir);
	} else {
		__dllPath->runtime_dir = absolutePath(__RUNTIME_DIR);
	}

	if (!tizenFXDir.empty()) {
		__dllPath->tizenfx_dir = absolutePath(tizenFXDir);
	} else {
		char* tmp = vconf_get_str(__TIZEN_API_PATH_KEY);
		if (tmp) {
			__dllPath->tizenfx_dir = std::string(tmp);
			_DBG("Device API Directory is set by vconf : %s", tmp);
			free(tmp);
		} else {
			__dllPath->tizenfx_dir = absolutePath(__DEVICE_API_DIR);
		}
	}

	__dllPath->tizenfx_ref_dir = __dllPath->tizenfx_dir + "/ref";

	// ":" seperated extra directories
	if (!extraDir.empty()) {
		splitPath(extraDir, __dllPath->extra_dirs);
	} else {
		char* extraPath = pluginGetDllPath();
		if (extraPath) {
			splitPath(extraPath, __dllPath->extra_dirs);
		}
	}

	_INFO("Path manager initialize success");

	initializedPathManager = true;
	return 0;
}

void finalizePathManager()
{
	if (__dllPath) {
		delete __dllPath;
		__dllPath = NULL;
	}

	initializedPathManager = false;
}

std::string getRuntimeDir()
{
	return __dllPath->runtime_dir;
}

std::string getTizenFXDir()
{
	return __dllPath->tizenfx_dir;
}

std::string getTizenFXRefDir()
{
	return __dllPath->tizenfx_ref_dir;
}

std::vector <std::string> getExtraDirs()
{
	return __dllPath->extra_dirs;
}

static std::string getPlatformTPA()
{
	std::string platform_tpa;

	if (isFileExist(PLATFORM_TPA_CACHE)) {
		_INFO("platform tpa cache found.");
		std::ifstream cacheFile;
		cacheFile.open(PLATFORM_TPA_CACHE);
		std::getline(cacheFile, platform_tpa);
		cacheFile.close();
	} else {
		std::vector<std::string> tpaDir;
		tpaDir.push_back(getRuntimeDir());
		tpaDir.push_back(getTizenFXDir());
		tpaDir.push_back(getTizenFXRefDir());
		assembliesInDirectory(tpaDir, platform_tpa);
	}

	return platform_tpa;
}

static std::string getPluginTPA()
{
	std::string plugin_tpa;

	char* plugin_tpa_list = pluginGetTPA();
	if (plugin_tpa_list) {
		_INFO("plugin TPA list found. use TPA list for plugin");
		plugin_tpa = plugin_tpa_list;
	} else if (!__dllPath->extra_dirs.empty()){
		_INFO("plugin extra directory found. use plugin extra directroy for TPA");
		assembliesInDirectory(__dllPath->extra_dirs, plugin_tpa);
	}

	return plugin_tpa;
}

std::string getTPA()
{
	if (!__tpa.empty()) {
		return __tpa;
	}

	if (__dllPath == NULL) {
		return std::string("");
	}

	return getPlatformTPA() + ":" + getPluginTPA();
}

